/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 by the Free Software Foundation.               *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "liberty.h"
#include "move_reasons.h"


/*
 * Private data structures used to collect information about moves.
 * FIXME: The static allocation should be replaced by some dynamic scheme.
 *        Also the data structures in general should probably be redesigned.
 */

#define MAX_MOVE_REASONS 1000
#define MAX_WORMS 2*MAX_BOARD*MAX_BOARD/3
#define MAX_DRAGONS MAX_WORMS
#define MAX_CONNECTIONS 4*MAX_WORMS
#define MAX_SEMEAIS MAX_WORMS
#define MAX_EYES MAX_BOARD*MAX_BOARD/2

static struct move_data move[MAX_BOARD][MAX_BOARD];
static struct move_reason move_reasons[MAX_MOVE_REASONS];
static int next_reason;

/* Worms */
static int wormi[MAX_WORMS];
static int wormj[MAX_WORMS];
static int next_worm;

/* Dragons */
static int dragoni[MAX_DRAGONS];
static int dragonj[MAX_DRAGONS];
static int next_dragon;

/* Connections */
static int conn_dragon1[MAX_CONNECTIONS];
static int conn_dragon2[MAX_CONNECTIONS];
static int next_connection;

/* Semeais */
static int semeai_myworm[MAX_SEMEAIS];
static int semeai_yourworm[MAX_SEMEAIS];
static int next_semeai;

/* Eye shapes */
static int eyei[MAX_EYES];
static int eyej[MAX_EYES];
static int eyecolor[MAX_EYES];
static int next_eye;

/* Initialize move reason data structures. */
void
clear_move_reasons(void)
{
  int i;
  int j;
  int k;
  next_reason = 0;
  next_worm = 0;
  next_dragon = 0;
  next_connection = 0;
  next_semeai = 0;
  
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      move[i][j].value = 0;
      move[i][j].shape_value = 0;
      move[i][j].old_style_value = 0;
      for (k=0; k<MAX_REASONS; k++)
	move[i][j].reason[k] = -1;
    }
}

/*
 * Find the index of a worm in the list of worms. If necessary,
 * add a new entry. (ai, aj) must point to the origin of the worm.
 */
static int
find_worm(int ai, int aj)
{
  int k;
  for (k=0; k<next_worm; k++)
    if ((wormi[k] == ai) && (wormj[k] == aj))
      return k;

  /* Add a new entry. */
  assert(next_worm < MAX_WORMS);
  wormi[next_worm] = ai;
  wormj[next_worm] = aj;
  next_worm++;
  return next_worm - 1;
}

/*
 * Find the index of a dragon in the list of dragons. If necessary,
 * add a new entry. (ai, aj) must point to the origin of the dragon.
 */
static int
find_dragon(int ai, int aj)
{
  int k;
  for (k=0; k<next_dragon; k++)
    if ((dragoni[k] == ai) && (dragonj[k] == aj))
      return k;

  /* Add a new entry. */
  assert(next_dragon < MAX_DRAGONS);
  dragoni[next_dragon] = ai;
  dragonj[next_dragon] = aj;
  next_dragon++;
  return next_dragon - 1;
}

/*
 * Find the index of a connection in the list of connections.
 * If necessary, add a new entry.
 */
static int
find_connection(int dragon1, int dragon2)
{
  int k;

  if (dragon1 > dragon2) {
    /* Swap to canonical order. */
    int tmp = dragon1;
    dragon1 = dragon2;
    dragon2 = tmp;
  }
  
  for (k=0; k<next_connection; k++)
    if ((conn_dragon1[k] == dragon1) && (conn_dragon2[k] == dragon2))
      return k;

  /* Add a new entry. */
  assert(next_connection < MAX_CONNECTIONS);
  conn_dragon1[next_connection] = dragon1;
  conn_dragon2[next_connection] = dragon2;
  next_connection++;
  return next_connection - 1;
}

/*
 * Find the index of a semeai in the list of semeais.
 * If necessary, add a new entry.
 */
static int
find_semeai(int myworm, int yourworm)
{
  int k;

  for (k=0; k<next_semeai; k++)
    if ((semeai_myworm[k] == myworm) && (semeai_yourworm[k] == yourworm))
      return k;

  /* Add a new entry. */
  assert(next_semeai < MAX_SEMEAIS);
  semeai_myworm[next_semeai] = myworm;
  semeai_yourworm[next_semeai] = yourworm;
  next_semeai++;
  return next_semeai - 1;
}

/*
 * Find the index of an eye space in the list of eye spaces.
 * If necessary, add a new entry.
 */
static int
find_eye(int i, int j, int color)
{
  int k;

  for (k=0; k<next_eye; k++)
    if ((eyei[k] == i) && (eyej[k] == j) && eyecolor[k] == color)
      return k;

  /* Add a new entry. */
  assert(next_eye < MAX_EYES);
  eyei[next_eye] = i;
  eyej[next_eye] = j;
  eyecolor[next_eye] = color;
  next_eye++;
  return next_eye - 1;
}

/*
 * Find a reason in the list of reasons. If necessary, add a new entry.
 */
static int
find_reason(int type, int what)
{
  int k;
  for (k=0; k<next_reason; k++)
    if ((move_reasons[k].type == type) && (move_reasons[k].what == what))
      return k;

  /* Add a new entry. */
  assert(next_reason < MAX_MOVE_REASONS);
  move_reasons[next_reason].type = type;
  move_reasons[next_reason].what = what;
  next_reason++;
  return next_reason - 1;
}

/*
 * Add a move reason for (ti, tj) if it's not already there or the
 * table is full.
 */ 
static void
add_move_reason(int ti, int tj, int type, int what)
{
  int k;
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[ti][tj].reason[k];
    if (r < 0)
      break;
    if (move_reasons[r].type == type
	&& move_reasons[r].what == what)
      return;  /* Reason already listed. */
  }
  /* Reason not found, add it if there is place left. */
  if (k<MAX_REASONS)
    move[ti][tj].reason[k] = find_reason(type, what);
}

/*
 * Remove a move reason for (ti, tj). Ignore silently if the reason
 * wasn't there.
 */ 
static void
remove_move_reason(int ti, int tj, int type, int what)
{
  int k;
  int n = -1; /* Position of the move reason to be deleted. */
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[ti][tj].reason[k];
    if (r < 0)
      break;
    if (move_reasons[r].type == type
	&& move_reasons[r].what == what)
      n = k;
  }

  if (n == -1)
    return; /* Move reason wasn't there. */

  /* Now move the last move reason to position n, thereby removing the
   * one we were looking for.
   */
  if (k == MAX_REASONS)
    k--;

  move[ti][tj].reason[n] = move[ti][tj].reason[k];
  move[ti][tj].reason[k] = -1;
}

/*
 * Add to the reasons for the move at (ti, tj) that it defends the worm
 * at (ai, aj).
 */
void
add_defense_move(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini,
				  worm[ai][aj].originj);
  add_move_reason(ti, tj, DEFEND_MOVE, worm_number);
}
  
/*
 * Add to the reasons for the move at (ti, tj) that it attacks the worm
 * at (ai, aj).
 */
void
add_attack_move(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  add_move_reason(ti, tj, ATTACK_MOVE, worm_number);
}

/*
 * Remove from the reasons for the move at (ti, tj) that it defends
 * the worm at (ai, aj). We do this by adding a NON_DEFEND move
 * reason and wait until later to actually remove it. Otherwise it may
 * be added again.
 */
void
remove_defense_move(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  add_move_reason(ti, tj, NON_DEFEND_MOVE, worm_number);
}


/*
 * Remove from the reasons for the move at (ti, tj) that it attacks
 * the worm at (ai, aj). We do this by adding a NON_ATTACK move
 * reason and wait until later to actually remove it. Otherwise it may
 * be added again.
 */
void
remove_attack_move(int ti, int tj, int ai, int aj)
{
  int worm_number = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  add_move_reason(ti, tj, NON_ATTACK_MOVE, worm_number);
}

/*
 * Add to the reasons for the move at (ti, tj) that it connects the
 * dragons at (ai, aj) and (bi, bj).
 */
void
add_connection_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  int dragon2 = find_dragon(dragon[bi][bj].origini, dragon[bi][bj].originj);
  int connection = find_connection(dragon1, dragon2);
  add_move_reason(ti, tj, CONNECT_MOVE, connection);
}

/*
 * Add to the reasons for the move at (ti, tj) that it cuts the
 * dragons at (ai, aj) and (bi, bj).
 */
void
add_cut_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  int dragon2 = find_dragon(dragon[bi][bj].origini, dragon[bi][bj].originj);
  int connection = find_connection(dragon1, dragon2);

  /*
   * Ignore the cut or connection if either (ai, aj) or (bi, bj)
   * points to a tactically captured worm.
   */
  if ((worm[ai][aj].attacki != -1 && worm[ai][aj].defendi == -1)
      || (worm[bi][bj].attacki != -1 && worm[bi][bj].defendi == -1))
    return;
  
  add_move_reason(ti, tj, CUT_MOVE, connection);
}

/*
 * Add to the reasons for the move at (ti, tj) that it is an anti-suji.
 * This means that it's a locally inferior move or for some other reason
 * must *not* be played.
 */
void
add_antisuji_move(int ti, int tj)
{
  add_move_reason(ti, tj, ANTISUJI_MOVE, 0);
}

/*
 * Add to the reasons for the move at (ti, tj) that it wins a semeai
 * between my worm at (ai, aj) and your worm at (bi, bj).
 */
void
add_semeai_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int myworm = find_worm(worm[ai][aj].origini, worm[ai][aj].originj);
  int yourworm = find_worm(worm[bi][bj].origini, worm[bi][bj].originj);
  int this_semeai = find_semeai(myworm, yourworm);
  add_move_reason(ti, tj, SEMEAI_MOVE, this_semeai);
}

/*
 * Add to the reasons for the move at (ti, tj) that it's the vital
 * point for the eye space at (ai, aj) of color color.
 */
void
add_vital_eye_move(int ti, int tj, int ai, int aj, int color)
{
  int eye;
  if (color == WHITE)
    eye = find_eye(white_eye[ai][aj].origini, white_eye[ai][aj].originj,
		   color);
  else
    eye = find_eye(black_eye[ai][aj].origini, black_eye[ai][aj].originj,
		   color);
  add_move_reason(ti, tj, VITAL_EYE_MOVE, eye);
}

/*
 * Increase or decrease the shape value for the move at (ti, tj).
 */
void
add_shape_value(int ti, int tj, int value)
{
  move[ti][tj].shape_value += value;
}

/*
 * Register a move proposed by the old patterns database.
 */
void
add_old_style_move_value(int ti, int tj, int value)
{
  if (move[ti][tj].old_style_value < value)
    move[ti][tj].old_style_value = value;
}

/*
 * Check whether a move reason already is recorded for a move.
 */
static int
move_reason_known(int ti, int tj, int type, int what)
{
  int k;
  int r;
  for (k=0; k<MAX_REASONS; k++) {
    r = move[ti][tj].reason[k];
    if (move_reasons[r].type == type
	&& move_reasons[r].what == what)
      return 1;
  }
  return 0;
}

/* Test all moves that defends, attacks, connects or cuts to see if
 * they also attack or defend some other worm.
 *
 * FIXME: Before starting to read, we should see if the considered move
 *        already is listed in the move_reasons.
 *
 * FIXME: We would like to see whether a move works to cut or connect
 *        something else too.
 */

static void
find_more_attack_and_defense_moves(int color)
{
  int unstable_worms[MAX_WORMS];
  int N = 0;  /* number of unstable worms */
  int m, n;
  int k;
  int other = OTHER_COLOR(color);
  
  /* Identify the unstable worms and store them in a list. */
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      if (p[m][n]
	  && worm[m][n].origini == m
	  && worm[m][n].originj == n
	  && worm[m][n].attacki != -1
	  && worm[m][n].defendi != -1) {
	unstable_worms[N] = find_worm(m, n);
	N++;
      }

  /* To avoid horizon effects, we temporarily increase the depth values. */
  increase_depth_values();
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      for (k=0; k<MAX_REASONS; k++) {
	int r = move[m][n].reason[k];
	if (r == -1
	    || move_reasons[r].type == ATTACK_MOVE
	    || move_reasons[r].type == DEFEND_MOVE
	    || move_reasons[r].type == CONNECT_MOVE
	    || move_reasons[r].type == CUT_MOVE)
	  break;
      }

      if (k<MAX_REASONS && move[m][n].reason[k] != -1) {
	/* Try the move at (m,n) and see what happens. */
	TRACE("trying %m\n",m ,n);
	if (trymove(m, n, color, "find_more_attack_and_defense_moves",
		    -1, -1)) {
	  for (k=0; k<N; k++) {
	    int ai = wormi[unstable_worms[k]];
	    int aj = wormj[unstable_worms[k]];
	    /* string of our color, see if there still is an attack,
	     * unless we already know the move works as defense move.
	     */
	    if (p[ai][aj] == color
		&& !move_reason_known(m, n, DEFEND_MOVE, unstable_worms[k]))
	      if (!attack(ai, aj, NULL, NULL)) {
		  TRACE("found extra point of defense of %m at %m\n",
			ai, aj, m, n);
		  add_defense_move(m, n, ai, aj);
	      }

	    /* string of opponent color, see if there still is a defense,
	     * unless we already know the move works as attack move.
	     */
	    if (p[ai][aj] == other
		&& !move_reason_known(m, n, ATTACK_MOVE, unstable_worms[k]))
	      if (!find_defense(ai, aj, NULL, NULL)) {
		/* Maybe find_defense() doesn't find the defense. Try to
		 * defend with the stored defense move.
		 */
		int attack_works = 1;
		if (trymove(worm[ai][aj].defendi, worm[ai][aj].defendj, other, 
			    "find_more_attack_and_defense_moves", -1, -1)) {
		  if (!attack(ai, aj, NULL, NULL))
		    attack_works = 0;
		  popgo();
		}
		
		if (attack_works) {
		  TRACE("found extra point of attack of %m at %m\n",
			ai, aj, m, n);
		  add_attack_move(m, n, ai, aj);
		}
	      }
	  }
	  popgo();
	}
      }
    }

  decrease_depth_values();
}

/*
 * Remove attacks and defenses that have earlier been marked as
 * NON_ATTACK or NON_DEFEND respectively, because they actually don't
 * work.
 */
static void
do_remove_false_attack_and_defense_moves(void)
{
  int m, n, k;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      for (k=0; k<MAX_REASONS; k++) {
	int r = move[m][n].reason[k];
	if (move_reasons[r].type == NON_ATTACK_MOVE)
	  remove_move_reason(m, n, ATTACK_MOVE, move_reasons[r].what);
	else if (move_reasons[r].type == NON_DEFEND_MOVE)
	  remove_move_reason(m, n, DEFEND_MOVE, move_reasons[r].what);
      }
    }
}

/*
 * Any move that captures or defends a worm also connects or cuts
 * the surrounding dragons. Find these secondary move reasons.
 *
 * FIXME: There is a certain amount of optimizations that could be
 *        done here.
 * FIXME: Even when we defend a worm, it's possible that the opponent
 *        still can secure a connection, e.g. underneath a string with
 *        few liberties. Thus a defense move isn't necessarily a cut move.
 */
static int adji[MAXCHAIN];
static int adjj[MAXCHAIN];
static int adjsize[MAXCHAIN];
static int adjlib[MAXCHAIN];

static void
induce_secondary_move_reasons(int color)
{
  int m;
  int n;
  int k;
  int i;
  int neighbors;
  int ai;
  int aj;
  int di = -1;
  int dj = -1;
  int best_strength = 0;
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      for (k=0; k<MAX_REASONS; k++) {

	int r = move[m][n].reason[k];
	if (r == -1)
	  break;

	if (move_reasons[r].type != ATTACK_MOVE
	    && move_reasons[r].type != DEFEND_MOVE)
	  continue;

	ai = wormi[move_reasons[r].what];
	aj = wormj[move_reasons[r].what];

	if ((   (move_reasons[r].type == ATTACK_MOVE)
		&& (p[ai][aj] == color))
	    || ((move_reasons[r].type == DEFEND_MOVE)
		&& (p[ai][aj] != color)))
	  continue; /* Only a move for the opponent. */
	
	if (worm[ai][aj].defendi == -1)
	  continue; /* No defense. */

	/*
	 * If this is a defense move and the defended worm is part of
	 * a dragon that is dead or has negative vitality, we only
	 * induce a cutting move if the same move also attacks or
	 * connects something or if the defended string has only one
	 * or two stones.
	 */
	if (dragon[ai][aj].size > 2
	    && move_reasons[r].type == DEFEND_MOVE
	    && (dragon[ai][aj].status == DEAD
		|| dragon[ai][aj].vitality < 0)) {
	  int ok = 0;
	  for (i=0; i<MAX_REASONS; i++) {
	    int rr = move[m][n].reason[i];
	    int type = move_reasons[rr].type;
	    int what = move_reasons[rr].what;
	    if (rr == -1)
	      break;
	    if ((type == ATTACK_MOVE
		 && p[wormi[what]][wormj[what]] != color)
		|| type == CONNECT_MOVE) {
	      ok = 1;
	      break;
	    }
	  }
	  if (!ok)
	    continue;
	}
	
	/* Find all dragons surrounding the worm. */
	chainlinks(ai, aj, &neighbors, adji, adjj, adjsize, adjlib);

	/*
	 * Find the strongest of the surrounding dragons and say that
	 * all other dragons are connected or cut with respect to that
	 * one.
	 *
	 * For the time being we use the vitality value of the dragon
	 * to measure strength. We probably want to change this later.
	 */
	for (i=0; i<neighbors; i++)
	  if (i == 0 || dragon[adji[i]][adjj[i]].vitality > best_strength) {
	    di = dragon[adji[i]][adjj[i]].origini;
	    dj = dragon[adji[i]][adjj[i]].originj;
	    best_strength = dragon[di][dj].vitality;
	  }

	for (i=0; i<neighbors; i++) {
	  int ei = dragon[adji[i]][adjj[i]].origini;
	  int ej = dragon[adji[i]][adjj[i]].originj;

	  if (di != ei || dj != ej) {
	    if (move_reasons[r].type == ATTACK_MOVE)
	      add_connection_move(m, n, di, dj, ei, ej);
	    else
	      add_cut_move(m, n, di, dj, ei, ej);
	  }
	}
      }
}

static void
list_move_reasons(int color)
{
  int m;
  int n;
  int k;
  int ai = -1;
  int aj = -1;
  int bi = -1;
  int bj = -1;
  int dragon1 = -1;
  int dragon2 = -1;
  int myworm = -1;
  int yourworm = -1;
  int ecolor = 0;
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      for (k=0; k<MAX_REASONS; k++) {

	int r = move[m][n].reason[k];
	if (r == -1)
	  break;

	switch(move_reasons[r].type) {
	  case ATTACK_MOVE:
	  case DEFEND_MOVE:
	    ai = wormi[move_reasons[r].what];
	    aj = wormj[move_reasons[r].what];

	    if (move_reasons[r].type == ATTACK_MOVE
		&& p[ai][aj] != color)
	      gprintf("Move at %m attacks %m%s\n", m, n, ai, aj,
		      (worm[ai][aj].defendi == -1) ? " (defenseless)" : "");
	    else if ((move_reasons[r].type == DEFEND_MOVE
		      && p[ai][aj] == color))
	      gprintf("Move at %m defends %m\n", m, n, ai, aj);
	    break;
	  case CONNECT_MOVE:
	  case CUT_MOVE:
	    dragon1 = conn_dragon1[move_reasons[r].what];
	    dragon2 = conn_dragon2[move_reasons[r].what];
	    ai = dragoni[dragon1];
	    aj = dragonj[dragon1];
	    bi = dragoni[dragon2];
	    bj = dragonj[dragon2];
	    if (move_reasons[r].type == CONNECT_MOVE)
	      gprintf("Move at %m connects %m and %m\n", m, n, ai, aj, bi, bj);
	    else
	      gprintf("Move at %m cuts %m and %m\n", m, n, ai, aj, bi, bj);
	    break;
	  case ANTISUJI_MOVE:
	    gprintf("Move at %m is an antisuji\n", m, n);
	    break;
	  case SEMEAI_MOVE:
	    myworm = semeai_myworm[move_reasons[r].what];
	    yourworm = semeai_yourworm[move_reasons[r].what];
	    ai = wormi[myworm];
	    aj = wormj[myworm];
	    bi = wormi[yourworm];
	    bj = wormj[yourworm];
	    gprintf("Move at %m urgent for semeai between %m and %m\n", m, n,
		    ai, aj, bi, bj);
	    break;
	  case VITAL_EYE_MOVE:
	    ai = eyei[move_reasons[r].what];
	    aj = eyej[move_reasons[r].what];
	    ecolor = eyecolor[move_reasons[r].what];
	    if (ecolor == WHITE)
	      gprintf("Move at %m vital eye point for dragon %m (eye %m)\n",
		      m, n, white_eye[ai][aj].dragoni,
		      white_eye[ai][aj].dragonj, ai, aj);
	    else
	      gprintf("Move at %m vital eye point for dragon %m (eye %m)\n",
		      m, n, black_eye[ai][aj].dragoni,
		      black_eye[ai][aj].dragonj, ai, aj);
	    break;
	}
      }
}

/*
 * Measure the "effective" size of a dragon. This measure is
 * completely ad hoc and not even reasonable.
 *
 * FIXME: Exchange this for something that is reasonable.
 */
static int effective_dragon_size(int ai, int aj)
{
#if 0
  int dragon_size = 0;
  int i;
  int j;

  /* Find all worms making up this dragon. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if (p[i][j] == p[ai][aj]
	  && worm[i][j].origini == i
	  && worm[i][j].originj == j
	  && dragon[i][j].origini == ai
	  && dragon[i][j].originj == aj)
	dragon_size += 2 * worm[ai][aj].size
	  + worm[ai][aj].liberties + worm[ai][aj].liberties2
	  + (worm[ai][aj].liberties3 + worm[ai][aj].liberties4)/2;
    }

  return dragon_size;
#endif
  return (int) (2 * dragon[ai][aj].effective_size);
}

/*
 * Strategical value of connecting (or cutting) the dragon at (ai, aj)
 * to the dragon at (bi, bj). Notice that this function is assymetric.
 *
 * FIXME: We shouldn't care about connecting killing nakade stones to
 *        the outside.
 */
static int
connection_value2(int ai, int aj, int bi, int bj)
{
  int status1 = dragon[ai][aj].status;
  int status2 = dragon[bi][bj].status;

  int dragon_size = effective_dragon_size(ai, aj);
  dragon_size /= (1 + dragon[ai][aj].escape_route / 4);

  /* An attempt to increase the importance of connections compared to the
   * old style patterns.
   */
  dragon_size *= 2;
  
  if (status1 == ALIVE) {
    if (dragon[ai][aj].vitality > 10)
      return 0;
    if (dragon[bi][bj].genus > 0 || dragon[bi][bj].heyes > 0
	|| status2 != DEAD)
      return dragon_size;
    else
      return 0;
  }
  
  if (status1 == CRITICAL) {
    if (status2 == ALIVE)
      return 6 * dragon_size;
    if (dragon[bi][bj].genus > 0 || dragon[bi][bj].heyes > 0
	|| status2 != DEAD)
      return 5 * dragon_size;
    else
      return 0;
  }
    
  if (status1 == UNKNOWN) {
    if (status2 == ALIVE) {
      if (dragon[ai][aj].vitality < 0)
	return 5 * dragon_size;
      else if (dragon[ai][aj].vitality < 10)
	return 3 * dragon_size;
      else
	return 2 * dragon_size;
    }
    if (dragon[bi][bj].genus > 0 || dragon[bi][bj].heyes > 0
	|| (status2 != DEAD && dragon[bi][bj].effective_size>4)) {
      if (dragon[ai][aj].vitality < 0)
	return 4 * dragon_size;
      else if (dragon[ai][aj].vitality < 10)
	return 2 * dragon_size;
      else
	return dragon_size;
    }
  }
    
  if (status1 == DEAD) {
    if (status2 == ALIVE)
      return 6 * dragon_size;
    if (2*dragon[ai][aj].genus + 2*dragon[bi][bj].genus
	+ dragon[ai][aj].heyes + dragon[bi][bj].heyes >=4)
      return 6 * dragon_size;
    if (status2 == UNKNOWN) {
      if (dragon[bi][bj].vitality < 0)
	return dragon_size;
      else if (dragon[bi][bj].vitality < 10)
	return 2 * dragon_size;
      else
	return 4 * dragon_size;
    }
  }

  return 0;
}

/*
 * Combine the reasons for a move at (m, n) into an old style value.
 * These heuristics are very ad hoc and not really expected to be very good.
 *
 * FIXME: Improve these.
 *
 */
static int
value_move_reasons(int m, int n, int color)
{
  int k;
  int ai = -1;
  int aj = -1;
  int bi = -1;
  int bj = -1;
  int dragon1 = -1;
  int dragon2 = -1;
  int myworm = -1;
  int yourworm = -1;
  int ecolor = 0;
  
  /* Strategical value of connecting or cutting dragons. */
  static int dragon_value[MAX_DRAGONS];

  int cutstone_bonus = 0; /* Do we attack or defend at least one cutstone or
			     potential cutstone? */
  int active_move = 0;    /* Are we doing something but just running away. */
  int tot_value = 0;
  int secondary_value = 0;
  for (k=0; k<next_dragon; k++)
    dragon_value[k] = 0;
  
  for (k=0; k<MAX_REASONS; k++) {
    int this_value = 0;
    int r = move[m][n].reason[k];
    if (r == -1)
      break;
    
    switch (move_reasons[r].type) {
      case ATTACK_MOVE:
      case DEFEND_MOVE:
	ai = wormi[move_reasons[r].what];
	aj = wormj[move_reasons[r].what];
	
	/* Attack on our stones. */
	if (move_reasons[r].type == ATTACK_MOVE
	    && p[ai][aj] == color) {
	  secondary_value += 1;
	  continue;
	}
	
	/* Defense of enemy stones. */
	if (move_reasons[r].type == DEFEND_MOVE
	    && p[ai][aj] != color) {
	  secondary_value += 1;
	  continue;
	}
	
	/* Defenseless stone */
	if (worm[ai][aj].defendi == -1) {
	  secondary_value += worm[ai][aj].size/2;
	  continue;
	}
	
	/* Territorial value. Strategical value is handled by cut/connection
	 * reasons and cutstone2 below.
	 */
	if (worm[ai][aj].liberties < 4
	    || (dragon[ai][aj].status != DEAD
		&& dragon[ai][aj].vitality > -5 )) {
	  this_value = (int) (4 * worm[ai][aj].effective_size);
	  DEBUG(DEBUG_MOVE_REASONS,"  %m: %d - worm %m\n",
		m, n, this_value, ai, aj);
	}
	
	if (worm[ai][aj].cutstone2 > 1) {
	  this_value += 50 * (worm[ai][aj].cutstone2 - 1);
	  DEBUG(DEBUG_MOVE_REASONS,"  %m: %d - %m cutstone\n", m, n, 
		50 * (worm[ai][aj].cutstone2 - 1), ai, aj);
	}

	if (worm[ai][aj].cutstone2 > 0)
	  cutstone_bonus = 1;

	tot_value += this_value;
	break;
	
      case CONNECT_MOVE:
      case CUT_MOVE:
	dragon1 = conn_dragon1[move_reasons[r].what];
	dragon2 = conn_dragon2[move_reasons[r].what];
	ai = dragoni[dragon1];
	aj = dragonj[dragon1];
	bi = dragoni[dragon2];
	bj = dragonj[dragon2];

	active_move = 1;
	
	this_value = connection_value2(ai, aj, bi, bj);
	if (this_value > dragon_value[dragon1])
	  dragon_value[dragon1] = this_value;
	
	this_value = connection_value2(bi, bj, ai, aj);
	if (this_value > dragon_value[dragon2])
	  dragon_value[dragon2] = this_value;
	
	break;
	
      case ANTISUJI_MOVE:
	/* This move must not be played. End of story. */
	return 0;
	break;
	
      case SEMEAI_MOVE:
	/*
	 * The territorial value of winning a semeai is about twice
	 * the sum of the effective sizes of the two worms plus other
	 * adjoining dead worms. If the alternative to winning is
	 * seki, the values are about half this size. The same goes
	 * for making seki instead of losing the semeai.
	 *
	 * The strategical value is that adjoining own dragons
	 * (usually) becomes fully secure, while adjoining enemy
	 * dragons do not.
	 *
	 * FIXME: Need to make a difference between the different
	 *        kinds of possible semeai results. There is also
	 *        some secondary value in playing an already won
	 *        semeai to eliminate ko threats, and vice versa.
	 *
	 * FIXME: Valuation not implemented at all yet.
	 */
	myworm = semeai_myworm[move_reasons[r].what];
	yourworm = semeai_yourworm[move_reasons[r].what];
	ai = wormi[myworm];
	aj = wormj[myworm];
	bi = wormi[yourworm];
	bj = wormj[yourworm];
	break;
	
      case VITAL_EYE_MOVE:
	ai = eyei[move_reasons[r].what];
	aj = eyej[move_reasons[r].what];
	ecolor = eyecolor[move_reasons[r].what];

	if (ecolor == WHITE) {
	  bi = white_eye[ai][aj].dragoni;
	  bj = white_eye[ai][aj].dragonj;
	}
	else {
	  bi = black_eye[ai][aj].dragoni;
	  bj = black_eye[ai][aj].dragonj;
	}

	if (bi == -1) /* Maybe we should assert this not to happen. */
	  break; 

	/* FIXME: These are ad hoc and need improvement. */
	if (dragon[bi][bj].status == CRITICAL) {
	  this_value = (int) (5 * dragon[bi][bj].effective_size);
	  DEBUG(DEBUG_MOVE_REASONS,"  %m: %d - vital for %m\n",
		m, n, this_value, bi, bj);
	  tot_value += this_value;
	}
	else
	  secondary_value += (int) (dragon[bi][bj].effective_size / 4);
	break;
    }
  }
  
  for (k=0; k<next_dragon; k++) {
    if (dragon_value[k] > 0)
      DEBUG(DEBUG_MOVE_REASONS,"  %m: %d - cuts and connections for %m\n",
	    m, n, dragon_value[k], dragoni[k], dragonj[k]);
    tot_value += dragon_value[k];
  }
  
  if (cutstone_bonus && active_move) {
    DEBUG(DEBUG_MOVE_REASONS,"  %m: 60 - cutstone bonus\n", m, n);
    tot_value += 60;
  }
  
  if (tot_value>0) {
    tot_value += (1+secondary_value)/2;
    if (secondary_value>1)
      DEBUG(DEBUG_MOVE_REASONS,"  %m: %d - secondary bonus\n", m, n,
	    (1+secondary_value)/2);
    tot_value += delta_moyo_simple(m, n, color);
    DEBUG(DEBUG_MOVE_REASONS,"  %m: %d - delta_moyo bonus\n", m, n,
	  delta_moyo_simple(m, n, color));
    if (move[m][n].shape_value != 0) {
      tot_value += move[m][n].shape_value;
      DEBUG(DEBUG_MOVE_REASONS,"  %m: %d - shape value\n", m, n,
	    move[m][n].shape_value);
    }
  }

  if (tot_value > 0) {
    TRACE("New move generation values %m to %d\n", 
	  m, n, tot_value);
    move_considered(m, n, tot_value);
  }

  /* The old style patterns may also benefit from a shape bonus,
   * but we should exclude endgame patterns, so we set a limit at 30.
   */
  {
    int old_value = move[m][n].old_style_value;
    if (old_value > 30)
      old_value += move[m][n].shape_value;
  
    if (tot_value < old_value) {
      if (tot_value > 0)
	TRACE("Old style value %d (%d + %d) dominates.\n", old_value,
	      move[m][n].old_style_value,
	      old_value - move[m][n].old_style_value);
      tot_value = old_value;
    }
  }
  
  return tot_value;
}

/*
 * Loop over all possible moves and value the move reasons for each.
 */
static void
value_moves(int color)
{
  int m;
  int n;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      move[m][n].value = value_move_reasons(m, n, color);
      if (move[m][n].value == 0.0)
	continue;
      /* Maybe this test should be performed elsewhere. This is just to get
       * some extra safety.
       */
      if (!legal(m, n, color)) {
	move[m][n].value = 0.0;
	DEBUG(DEBUG_MOVE_REASONS, "Move at %m wasn't legal.\n", m, n);
      }
      else {
	/* Add a random number between 0 and 0.5 to use in comparisons. */
	/* FIXME: Find a proper float valued rand function. Need to
	 *        make sure it's portable and/or autoconfed.
	 */
	move[m][n].value += (rand()%5000)/10000.0;
      }
    }
}

/* Replacement for attacker() and defender(). Also handles
 * cuts/connections and eye-finding.
 *
 * FIXME: This function is increasingly misleadingly named.
 */
int
experimental_attdef(int *i, int *j, int *val, int color)
{
  int m;
  int n;
  float tval;
  float bestval = 0.0;

  find_more_attack_and_defense_moves(color);
  do_remove_false_attack_and_defense_moves();
  induce_secondary_move_reasons(color);

  if (verbose)
    list_move_reasons(color);
  
  value_moves(color);
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if (move[m][n].value == 0.0)
	continue;

      tval = move[m][n].value;

      if (tval > bestval) {
	bestval = tval;
	*i = m;
	*j = n;
      }
    }
  
  if (bestval > 0.0) {
    *val = (int) bestval;
    return 1;
  }
  return 0;
}

/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */

